"""FakeAtoms and FakeSurfaceCalculator to run ASE optimisations and dynamics within an N-dimensional PMF/energy surface"""

from .hijack import FakeAtoms, FakeSurfaceCalculator, NaNEncounteredInPmfError
