standard = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": False,
    "radii": 0.9,
    "rotation": "",
    "celllinewidth": 0.0,
}

standard_rot = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": False,
    "radii": 0.9,
    "rotation": "-90x,65z",
    "celllinewidth": 0.0,
}

standard_bonds = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": False,
    "radii": 0.5,
    "celllinewidth": 0.0,
}

standard_cell = {
    "canvas_width": 2000,
    "canvas_height": 2000,
    "drawCell": True,
    "radii": 0.9,
    "celllinewidth": 0.07,
}

gif_standard = {"loop": 0, "fps": 25, "subrectangles": True}
